/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file comparison.c
 * @brief This file contains the function used for the comparison of the
 * filtering methods of Algorithm 1. See the function comparison_all_filters
 * for more details.
 * @version 1.0
 * @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
 * @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
 */

#include <stdlib.h>
#include <math.h>
#include "filtering.h"
#include "filters.h"
#include "toolbox.h"

#define FORI(n) for(int i=0;i<(n);i++)
#define FORJ(n) for(int j=0;j<(n);j++)
#define FORL(n) for(int l=0;l<(n);l++)

/* Image spectral content */

/**
 * @brief Compute the boundary value and total value (see Definition ?)
 * of an image.
 * @param boundary computed boundary value
 * @param total computed total value
 * @param inhat input DFT array
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
static void compute_boundary_value(float *boundary, float *total, float *inhat,
                      int w, int h, int pd)
{
  /* total value */
  *total = 0;
  FORL(pd){
    FORI(w*h) *total += hypot(inhat[2*i*pd+2*l],inhat[2*i*pd+2*l+1]);
  }
  *total = *total/(pd*w*h); // scaling

  /* boundary value */
  float (*inhat2)[w][2*pd] = (void*) inhat;
  int mw = w/2;
  int mh = h/2;

  *boundary = 0;
  /* if the widht is an even number */
  if(2*mw == w){
      FORL(pd){
        FORJ(h) *boundary += hypot(inhat2[j][mw][2*l],inhat2[j][mw][2*l+1]);
      }
  }
  /* if the height is an even number */
  if(2*mh == h){
    FORL(pd){
      FORI(w) *boundary += hypot(inhat2[mh][i][2*l],inhat2[mh][i][2*l+1]);
     }
  }
  /* if both sizes are even we have summed two times the same term */
  if(2*mw == w && 2*mh == h)
     FORL(pd) *boundary -= hypot(inhat2[mh][mw][2*l],inhat2[mh][mw][2*l+1]);
  /* rescaling */
  *boundary = *boundary/(pd*w*h);
}

/**
 * @brief Compute the modulus of a DFT array in logarithmic scale
 * (see Section 4.2.1).
 * The output is rearranged so that the zero-frequency component is
 * in the center of the array (fftshift).
 * @param out computed modulus of the DFT input in logarithmic scale
 * @param inhat input DFT array
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
static void compute_mod_fft(float *out, float *inhat, int w, int h, int pd)
{
  float *out_shifted = malloc(w*h*pd*sizeof*out_shifted);
  float (*out2)[w][pd] = (void*) out;
  float (*out_shifted2)[w][pd] = (void*) out_shifted;

  /* compute modulus of the input (logarithmic scale) */
  FORL(pd){
    FORI(w*h) out_shifted[i*pd+l] = log(1+hypot(inhat[2*(i*pd + l)+0],
                                                inhat[2*(i*pd + l)+1]));
  }

  /* Rearrangement */
  int w_test = w/2;
  int h_test = h/2;
  int cw = (2*w_test == w) ? 1 : 0;
  int ch = (2*h_test == h) ? 1 : 0;
  int ind_w = (w-1)/2;
  int ind_h = (h-1)/2;
  int i2,j2;
  FORL(pd){
    FORI(w){
      i2 = (i<ind_w+1) ? i+ind_w+cw : i-ind_w-1;
      FORJ(h){
        j2 = (j<ind_h+1) ? j+ind_h+ch : j-ind_h-1;
        out2[j2][i2][l] = out_shifted2[j][i][l];
      }
    }
  }

  /* free memory */
  free(out_shifted);
}

/**
 * @brief This function computes for two filtered images the difference image,
 * its mean and its maximum value. See Section 4.2.2.
 * @param diff computed difference image
 * @param im1 input image 1
 * @param im2 input image 2
 * @param mean computed mean
 * @param diff_max computed maximal difference
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
static void compare_two_methods(float *diff, float *im1, float *im2,
                                float *mean, float *diff_max,
                                int w, int h, int pd)
{
  float diff_preal, diff_pcomplex;
  *mean = 0;

  /* compute the difference image and the sum of its elements */
  FORL(pd){
    FORI(w*h){
      diff_preal = im1[2*(i*pd + l)+0]-im2[2*(i*pd + l)+0];
      diff_pcomplex = im1[2*(i*pd + l)+1]-im2[2*(i*pd + l)+1];
      diff[i*pd+l] = hypot(diff_preal,diff_pcomplex);
      *mean += diff[i*pd+l];
    }
  }
  /* renormalization of the sum to get the mean */
  *mean = *mean/(w*h*pd);
  /* maximum value */
  *diff_max = max(diff,w*h*pd);
}

/**
 * @brief This function compares the three filtering methods for a given filter.
 * It first computes the three filtered images. Then for each possible
 * pair (1-2, 1-3, 2-3) it compares the filtered images by computing their
 * difference image, its mean and its maximal value. It also computes the range
 * of the real part of the filtered image with method 1 (this normalization value
 * is used when the difference image is saved).
 * Finally it computes :
 * - 3 filtered images
 * - 3 difference images
 * - 6 values (3 mean + 3 max)
 * - 1 range
 * @param diff computed difference image
 * @param im1 input image 1
 * @param im2 input image 2
 * @param mean computed mean
 * @param diff_max computed maximal difference
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
static void comparison_one_filter(float **filtered_images, float **diff_images,
                           float *print_info, float *range_value, filter_t f,
                           float *in, int w, int h, int pd)
{
  /* compute the filtered images with the three methods */
  for(int method = 1; method<=3; method++)
     apply_filter(*(filtered_images+method-1), in, f, method,
                  w, h, pd);

  /* compare the filtered images */
    /* 1 v 2 */
    compare_two_methods(*diff_images, *filtered_images,
                        *(filtered_images+1), print_info, print_info+1,
                        w, h, pd);
    /* 1 v 3 */
    compare_two_methods(*(diff_images+1), *filtered_images,
                        *(filtered_images+2), print_info+2, print_info+3,
                        w, h, pd);
    /* 2 v 3 */
    compare_two_methods(*(diff_images+2), *(filtered_images+1),
                        *(filtered_images+2), print_info+4, print_info+5,
                        w, h, pd);

  /* compute the range of the real part of the filtered image with method 1 */
     float *real_part_for_range = malloc(w*h*pd*sizeof(float));
     real_part(real_part_for_range, *filtered_images, w, h, pd);
     *range_value = range(real_part_for_range, w*h*pd);
}

/**
 * @brief This function compares the three filtering methods of Algorithm 1 for
 * 9 proposed filters (see Section 4.2.3). It first computes the boundary
 * value, the total value and the dft of the input image.
 * For each filter it computes :
 * - 3 filtered images
 * - 3 difference images
 * - 6 values (3 mean + 3 mean)
 * - 1 range
 * Finally it computes 27 filtered images and 27 difference images
 * indexed as follows :
 * 0-1-2    : sinc filter
 * 3-4-5    : shift filter
 * 6-7-8    : derivative filter
 * 9-10-11  : laplacian filter
 * 12-13-14 : gaussian filter
 * 15-16-17 : lowpass filter
 * 18-19-20 : highpass filter
 * 21-22-23 : horizontal filter
 * 24-25-26 : diagonal filter
 * The 56 computed values are stored with the following indices in print_info :
 * 0     : boundary value
 * 1     : total value
 * 2-7   : sinc values
 * 8-13  : shift values
 * 14-19 : derivative values
 * 20-25 : laplacian values
 * 26-31 : gaussian values
 * 32-37 : lowpass values
 * 38-43 : highpass values
 * 44-49 : horizontal values
 * 50-55 : diagonal values
 * The 9 range values are stored with the following indices :
 * 0 : sinc filter
 * 1 : shift filter
 * 2 : derivative filter
 * 3 : laplacian filter
 * 4 : gaussian filter
 * 5 : lowpass filter
 * 6 : highpass filter
 * 7 : horizontal filter
 * 8 : diagonal filter
 * @param filtered_images the filtered images (complex-valued)
 * @param diff_images the difference images (real-valued)
 * @param modulus_dft_in the modulus of the DFT input in logarithmic scale
 * @param print_info computed values (that are printed)
 * @param range_value computed range values
 * @param in input array
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
void comparison_all_filters(float **filtered_images, float **diff_images,
                       float *modulus_dft_in, float *print_info,
                       float *range_value, float *in, int w, int h, int pd)
{
  /* image properties */
    /* compute the DFT input */
    float *inhat=malloc(2*w*h*pd*sizeof(float));
    fft_direct(inhat, in, w, h, pd);

    /* boundary energy of the image */
    compute_boundary_value(&print_info[0], &print_info[1], inhat, w, h, pd);

    /* compute the modulus (log scale) of the DFT */
    compute_mod_fft(modulus_dft_in, inhat, w, h, pd);

  /* Comparison of the methods */
  filter_t f;
  int ind1 = 0; // for filtered_images and diff_images indices
  int ind2 = 2; // for print_info indices
  int ind3 = 0; // for range_value indices

    /* sinc filter */
    f = sinc_filter;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* shift filter */
    f = shift_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* derivative filter */
    f = derivative_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* laplacian filter */
    f = laplacian_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* gaussian filter */
    f = gaussian_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* lowpass filter */
    f = lowpass_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* highpass filter */
    f = highpass_filter;
    ind1 += 3;
    ind2 += 6;
    ind3 += 1;
    comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                          &range_value[ind3], f, in, w, h, pd);

    /* steered filters */
    f = steered_filter;
    global_alpha = compute_alpha(global_number_steer);
      /* Horizontal filter */
      global_steer = 0;
      ind1 += 3;
      ind2 += 6;
      ind3 += 1;
      comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                            &range_value[ind3], f, in, w, h, pd);

      /* Horizontal filter */
      global_steer = global_number_steer/4;
      ind1 += 3;
      ind2 += 6;
      ind3 += 1;
      comparison_one_filter(filtered_images+ind1, diff_images+ind1, print_info+ind2,
                            &range_value[ind3], f, in, w, h, pd);

  /* free memory */
  free(inhat);
}
