/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
* @file filtering.c
* @brief This file contains the implementation of Algorithm 1 (see Section 3).
* It is done in the function apply_filter. The filtered image, which may be
* complex-valued, is stored in a float array with twice the size of the input
* image. The function real_part_apply_filter calls the function apply_filter
* and then takes the real part of the filtered image.
* @version 1.0
* @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
* @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
*/

#include <stdlib.h>
#include "filters.h"
#include "toolbox.h"

/* M_PI is a POSIX definition */
#ifndef M_PI
/** macro definition for Pi */
#define M_PI 3.14159265358979323846
#endif
/* !M_PI */

#define FORI(n) for(int i=0;i<(n);i++)
#define FORJ(n) for(int j=0;j<(n);j++)
#define FORL(n) for(int l=0;l<(n);l++)

/* typedef to handle the name of the function that defines the filter */
typedef complex float (*filter_t)(float, float);

/**
 * @brief This function computes the (spectral) coordinates corresponding
 * to array indexes.
 * @param x x-coordinate corresponding to i (output value)
 * @param y y-coordinate corresponding to j (output value)
 * @param i widht index
 * @param j height index
 * @param w input width
 * @param h input height
 */
static void pixel_location(float *x, float *y, int i, int j, int w, int h)
{
  /* x component */
  int ind_w = (w-1)/2;
  *x = (i<ind_w+1) ? i*2*M_PI/(w): (i-w)*2*M_PI/(w);
  /* y component */
  int ind_h = (h-1)/2;
  *y = (j<ind_h+1) ? j*2*M_PI/(h) : (j-h)*2*M_PI/(h);
}

/**
 * @brief This function computes the spectral samples corresponding to
 * a function defines in the Fourier domain, the given method and a size
 * of image (step 2 of Algorithm 1).
 * The function first computes the spectral samples of method 1. The
 * boundary is then modified if method=2 or method=3.
 * @param spectral_samples float array of the computed spectral samples
 * @param f function that defines the filter in the Fourier domain
 * @param method method of the filtering algorithm (1,2 or 3)
 * @param w input width
 * @param h input height
 */
static void compute_spectral_samples(float *spectral_samples, filter_t f, int method,
                              int w, int h)
{
  /* declaration */
  float (*spectral_samples2)[w][2] = (void*) spectral_samples;
  float x, y;
  complex float f1, f2, f3, f4;

  /* boundary indices */
  int mw = w/2;
  int mh = h/2;

  /* Fill the output array with the spectral samples of method 1 */
     FORJ(h){
       FORI(w){
         pixel_location(&x, &y, i, j, w, h);
         f1=f(x,y);
         spectral_samples2[j][i][0] = creal(f1);
         spectral_samples2[j][i][1] = cimag(f1);
       }
     }

  if( method == 2 ){
    /* modify the boundary */
    if( 2*mw == w ){
      FORJ(h){
        pixel_location(&x, &y, mw, j, w, h);
        f1=f(x,y);
        f2=f(-x, y);
        spectral_samples2[j][mw][0] = 0.5*(creal(f1+f2));
        spectral_samples2[j][mw][1] = 0.5*(cimag(f1+f2));
      }
    }
    if( 2*mh == h ){
      FORI(w){
        pixel_location(&x, &y, i, mh, w, h);
        f1=f(x,y);
        f2=f(x,-y);
        spectral_samples2[mh][i][0] = 0.5*(creal(f1+f2));
        spectral_samples2[mh][i][1] = 0.5*(cimag(f1+f2));
      }
    }
    if( 2*mw == w && 2*mh == h )
    {
      x = M_PI;
      y = M_PI;
      f1=f(x,y);
      f2=f(-x,y);
      f3=f(x,-y);
      f4=f(-x,-y);
      spectral_samples2[mh][mw][0] = 0.25*(creal(f1+f2+f3+f4));
      spectral_samples2[mh][mw][1] = 0.25*(cimag(f1+f2+f3+f4));
    }
  }

  if( method == 3 ){
    /* Set the boundary to zero */
    if( 2*mw == w ){
      FORJ(h){
        spectral_samples2[j][mw][0] = 0;
        spectral_samples2[j][mw][1] = 0;
      }
    }
    if( 2*mh == h ){
      FORI(w){
        spectral_samples2[mh][i][0] = 0;
        spectral_samples2[mh][i][1] = 0;
      }
    }
  }
}

/**
 * @brief This function computes element_wise_multiplication of a DFT array
 * and spectral samples (step 3 of Algorithm 1).
 * The DFT array may have several channels while the
 * spectral samples only have one channel. Each channel of the DFT is
 * multiplied by the spectral samples.
 * @param outhat array of element-wise multiplication
 * @param inhat input DFT array
 * @param spectral_samples spectral samples array
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
static void element_wise_multiplication(float *outhat, float *inhat,
                                        float *spectral_samples,
                                        int w, int h, int pd)
{
  /* trick for indices correspondances */
  float (*inhat2)[w][2*pd] = (void*) inhat;
  float (*spectral_samples2)[w][2] = (void*) spectral_samples;
  float (*outhat2)[w][2*pd] = (void*) outhat;

  /* element-wise multiplication */
  FORI(w){
    FORJ(h){
      FORL(pd){
        outhat2[j][i][2*l] = inhat2[j][i][2*l]*spectral_samples2[j][i][0]
                             -inhat2[j][i][2*l+1]*spectral_samples2[j][i][1];
        outhat2[j][i][2*l+1] = inhat2[j][i][2*l]*spectral_samples2[j][i][1]
                               +inhat2[j][i][2*l+1]*spectral_samples2[j][i][0];
      }
    }
  }
}

/**
 * @brief This function is an implementation of Algorithm 1 (see Section 3).
 * The output image which is complex-valued is stored in a float array with
 * twice the size of the input array.
 * @param out filtered image (complex-valued)
 * @param in input image
 * @param f function that defines the filter in the Fourier domain
 * @param method method of the filtering algorithm (1,2 or 3)
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
void apply_filter(float *out, float *in, filter_t f, int method,
                  int w, int h, int pd)
{
  /* memory allocation */
  float *inhat = malloc(2*w*h*pd*sizeof(float));
  float *outhat = malloc(2*w*h*pd*sizeof(float));
  float *spectral_samples = malloc(2*w*h*sizeof(float));

  /* Step 1 : compute the DFT input */
  fft_direct(inhat, in, w, h, pd);

  /* Step 2 : compute the spectral samples */
  compute_spectral_samples(spectral_samples, f, method, w, h);

  /* Step 3 : element-wise multiplication */
  element_wise_multiplication(outhat, inhat, spectral_samples,
                              w, h, pd);

  /* Step 4 : compute the filtered image by iDFT */
  fft_inverse(out, outhat, w, h, pd);

  /* free memory */
  free(inhat);
  free(outhat);
  free(spectral_samples);
}

/**
 * @brief This function computes the real part of the filtered image
 * obtained with Algorithm 1. It calls the function apply_filter
 * and then take the real part of its output.
 * @param out real part of the filtered image
 * @param in input image
 * @param f function that defines the filter in the Fourier domain
 * @param method method of the filtering algorithm (1,2 or 3)
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
void real_part_apply_filter(float *out, float *in, filter_t f, int method,
                            int w, int h, int pd)
{
  /* memory allocation */
  float *out_complex = malloc(2*w*h*pd*sizeof(float));

  /* compute the filtered image */
  apply_filter(out_complex, in, f, method, w, h, pd);
  real_part(out, out_complex, w, h, pd);

  /* free memory */
  free(out_complex);
}
